# AI 전문가 양성 과정 - 환경 설정 자동화 스크립트 (Windows PowerShell)
# 이 스크립트는 필요한 Python 패키지들을 자동으로 설치합니다.

# 오류 발생 시 스크립트 중단
$ErrorActionPreference = "Stop"

Write-Host "🚀 AI 전문가 양성 과정 환경 설정을 시작합니다..." -ForegroundColor Green
Write-Host "==========================================" -ForegroundColor Green

# Python 버전 확인
function Check-Python {
    Write-Host "Python 버전을 확인합니다..." -ForegroundColor Blue
    
    try {
        $pythonVersion = python --version 2>&1
        if ($pythonVersion -match "Python (\d+\.\d+\.\d+)") {
            $version = $matches[1]
            Write-Host "Python 버전: $version" -ForegroundColor Green
            
            # Python 3.8 이상 필요
            $versionParts = $version.Split('.')
            $major = [int]$versionParts[0]
            $minor = [int]$versionParts[1]
            
            if ($major -ge 3 -and $minor -ge 8) {
                Write-Host "✓ Python 버전이 요구사항을 만족합니다." -ForegroundColor Green
            } else {
                Write-Host "✗ Python 3.8 이상이 필요합니다. 현재 버전: $version" -ForegroundColor Red
                exit 1
            }
        } else {
            throw "Python 버전을 확인할 수 없습니다."
        }
    } catch {
        Write-Host "✗ Python이 설치되어 있지 않거나 PATH에 추가되지 않았습니다." -ForegroundColor Red
        Write-Host "Python을 먼저 설치하고 PATH에 추가해주세요." -ForegroundColor Yellow
        Write-Host "다운로드: https://www.python.org/downloads/" -ForegroundColor Yellow
        exit 1
    }
}

# pip 업그레이드
function Upgrade-Pip {
    Write-Host "pip를 최신 버전으로 업그레이드합니다..." -ForegroundColor Blue
    python -m pip install --upgrade pip
    Write-Host "✓ pip 업그레이드 완료" -ForegroundColor Green
}

# 가상환경 생성 및 활성화
function Setup-VirtualEnvironment {
    Write-Host "가상환경을 설정합니다..." -ForegroundColor Blue
    
    if (-not (Test-Path "venv")) {
        Write-Host "가상환경을 생성합니다..." -ForegroundColor Yellow
        python -m venv venv
        Write-Host "✓ 가상환경 생성 완료" -ForegroundColor Green
    } else {
        Write-Host "가상환경이 이미 존재합니다." -ForegroundColor Yellow
    }
    
    Write-Host "가상환경을 활성화합니다..." -ForegroundColor Yellow
    & ".\venv\Scripts\Activate.ps1"
    Write-Host "✓ 가상환경 활성화 완료" -ForegroundColor Green
}

# 기본 패키지 설치
function Install-BasicPackages {
    Write-Host "기본 패키지들을 설치합니다..." -ForegroundColor Blue
    
    python -m pip install --upgrade pip setuptools wheel
    
    # requirements.txt가 있다면 사용, 없다면 기본 패키지 설치
    if (Test-Path "requirements.txt") {
        Write-Host "requirements.txt에서 패키지를 설치합니다..." -ForegroundColor Yellow
        pip install -r requirements.txt
    } else {
        Write-Host "기본 패키지들을 설치합니다..." -ForegroundColor Yellow
        $packages = @(
            "numpy>=1.21.0",
            "pandas>=1.3.0", 
            "matplotlib>=3.4.0",
            "seaborn>=0.11.0",
            "scikit-learn>=1.0.0",
            "jupyter>=1.0.0",
            "notebook>=6.4.0",
            "ipykernel>=6.0.0"
        )
        
        foreach ($package in $packages) {
            pip install $package
        }
    }
    
    Write-Host "✓ 기본 패키지 설치 완료" -ForegroundColor Green
}

# 딥러닝 패키지 설치 (선택사항)
function Install-DeepLearningPackages {
    $response = Read-Host "딥러닝 패키지를 설치하시겠습니까? (y/N)"
    
    if ($response -eq "y" -or $response -eq "Y") {
        Write-Host "딥러닝 패키지들을 설치합니다..." -ForegroundColor Blue
        
        # CPU 버전 PyTorch 설치
        pip install torch torchvision torchaudio --index-url https://download.pytorch.org/whl/cpu
        
        # TensorFlow 설치
        pip install "tensorflow>=2.8.0"
        
        # 추가 딥러닝 라이브러리
        $dlPackages = @(
            "transformers>=4.15.0",
            "datasets>=2.0.0", 
            "torch-geometric",
            "opencv-python>=4.5.0"
        )
        
        foreach ($package in $dlPackages) {
            pip install $package
        }
        
        Write-Host "✓ 딥러닝 패키지 설치 완료" -ForegroundColor Green
    } else {
        Write-Host "딥러닝 패키지 설치를 건너뜁니다." -ForegroundColor Yellow
    }
}

# NLP 패키지 설치 (선택사항)
function Install-NLPPackages {
    $response = Read-Host "NLP 패키지를 설치하시겠습니까? (y/N)"
    
    if ($response -eq "y" -or $response -eq "Y") {
        Write-Host "NLP 패키지들을 설치합니다..." -ForegroundColor Blue
        
        $nlpPackages = @(
            "nltk>=3.7",
            "spacy>=3.4.0",
            "langchain>=0.0.200",
            "openai>=0.27.0",
            "sentence-transformers>=2.2.0"
        )
        
        foreach ($package in $nlpPackages) {
            pip install $package
        }
        
        # spaCy 모델 다운로드
        python -m spacy download en_core_web_sm
        
        Write-Host "✓ NLP 패키지 설치 완료" -ForegroundColor Green
    } else {
        Write-Host "NLP 패키지 설치를 건너뜁니다." -ForegroundColor Yellow
    }
}

# Jupyter 커널 등록
function Setup-JupyterKernel {
    Write-Host "Jupyter 커널을 등록합니다..." -ForegroundColor Blue
    
    python -m ipykernel install --user --name=ai_course --display-name="AI Course"
    
    Write-Host "✓ Jupyter 커널 등록 완료" -ForegroundColor Green
}

# 개발 도구 설치 (선택사항)
function Install-DevTools {
    $response = Read-Host "개발 도구를 설치하시겠습니까? (코드 포맷팅, 린팅 등) (y/N)"
    
    if ($response -eq "y" -or $response -eq "Y") {
        Write-Host "개발 도구들을 설치합니다..." -ForegroundColor Blue
        
        $devPackages = @(
            "black>=22.0.0",
            "flake8>=4.0.0",
            "isort>=5.10.0",
            "mypy>=0.950",
            "pytest>=7.0.0",
            "pre-commit>=2.17.0"
        )
        
        foreach ($package in $devPackages) {
            pip install $package
        }
        
        Write-Host "✓ 개발 도구 설치 완료" -ForegroundColor Green
    } else {
        Write-Host "개발 도구 설치를 건너뜁니다." -ForegroundColor Yellow
    }
}

# 설치 완료 확인
function Verify-Installation {
    Write-Host "설치된 패키지를 확인합니다..." -ForegroundColor Blue
    
    $verifyScript = @"
import sys
print(f'Python 버전: {sys.version}')

packages = ['numpy', 'pandas', 'matplotlib', 'seaborn', 'sklearn', 'jupyter']

for package in packages:
    try:
        __import__(package)
        print(f'✓ {package}: 설치됨')
    except ImportError:
        print(f'✗ {package}: 설치되지 않음')
"@
    
    python -c $verifyScript
    
    Write-Host "✓ 패키지 확인 완료" -ForegroundColor Green
}

# 사용법 안내
function Show-Usage {
    Write-Host "==========================================" -ForegroundColor Blue
    Write-Host "🎉 환경 설정이 완료되었습니다!" -ForegroundColor Blue
    Write-Host "==========================================" -ForegroundColor Blue
    Write-Host ""
    Write-Host "다음 명령어로 가상환경을 활성화하세요:" -ForegroundColor Yellow
    Write-Host ".\venv\Scripts\Activate.ps1" -ForegroundColor White
    Write-Host ""
    Write-Host "Jupyter Notebook을 시작하려면:" -ForegroundColor Yellow
    Write-Host "jupyter notebook" -ForegroundColor White
    Write-Host ""
    Write-Host "코드를 실행하려면:" -ForegroundColor Yellow
    Write-Host "cd source_code\02_python_core_syntax" -ForegroundColor White
    Write-Host "python part_2_python_core_syntax.py" -ForegroundColor White
    Write-Host ""
    Write-Host "문제가 발생하면 다음을 확인해주세요:" -ForegroundColor Yellow
    Write-Host "- Python 버전이 3.8 이상인지 확인" -ForegroundColor White
    Write-Host "- 가상환경이 활성화되어 있는지 확인" -ForegroundColor White
    Write-Host "- 인터넷 연결 확인" -ForegroundColor White
    Write-Host "- Windows에서 실행 정책 확인: Set-ExecutionPolicy -ExecutionPolicy RemoteSigned" -ForegroundColor White
}

# 메인 실행 함수
function Main {
    Write-Host "AI 전문가 양성 과정 환경 설정" -ForegroundColor Green
    Write-Host ""
    
    try {
        Check-Python
        Upgrade-Pip
        Setup-VirtualEnvironment
        Install-BasicPackages
        Install-DeepLearningPackages
        Install-NLPPackages
        Setup-JupyterKernel
        Install-DevTools
        Verify-Installation
        Show-Usage
        
        Write-Host "🎉 모든 설정이 완료되었습니다!" -ForegroundColor Green
    } catch {
        Write-Host "오류가 발생했습니다: $($_.Exception.Message)" -ForegroundColor Red
        Write-Host "스크립트를 다시 실행하거나 수동으로 설치해주세요." -ForegroundColor Yellow
        exit 1
    }
}

# 실행 정책 확인
if ((Get-ExecutionPolicy) -eq "Restricted") {
    Write-Host "PowerShell 실행 정책이 제한되어 있습니다." -ForegroundColor Yellow
    Write-Host "다음 명령어를 관리자 권한으로 실행하세요:" -ForegroundColor Yellow
    Write-Host "Set-ExecutionPolicy -ExecutionPolicy RemoteSigned" -ForegroundColor White
    Write-Host ""
    $response = Read-Host "계속하시겠습니까? (y/N)"
    if ($response -ne "y" -and $response -ne "Y") {
        exit 1
    }
}

# 스크립트 실행
Main 